import inspect
import json
from contextlib import suppress
from functools import partial
from typing import Any, Optional, Tuple, Union, cast, overload

from faststream._compat import dump_json
from faststream.broker.message import StreamMessage
from faststream.broker.types import (
    AsyncCustomDecoder,
    AsyncCustomParser,
    AsyncDecoder,
    AsyncParser,
    CustomDecoder,
    CustomParser,
    Decoder,
    MsgType,
    Parser,
    SyncCustomDecoder,
    SyncCustomParser,
    SyncDecoder,
    SyncParser,
)
from faststream.constants import ContentType, ContentTypes
from faststream.types import DecodedMessage, SendableMessage


def decode_message(message: StreamMessage[Any]) -> DecodedMessage:
    """Decodes a message.

    Args:
        message: The message to decode.

    Returns:
        The decoded message.

    Raises:
        JSONDecodeError: If the message body cannot be decoded as JSON.
    !!! note

        The above docstring is autogenerated by docstring-gen library (https://docstring-gen.airt.ai)
    """
    body = message.body
    m: DecodedMessage = body

    if message.content_type:
        if ContentTypes.text.value in message.content_type:
            m = body.decode()
        elif ContentTypes.json.value in message.content_type:  # pragma: no branch
            m = json.loads(body)

    else:
        with suppress(json.JSONDecodeError):
            m = json.loads(body)

    return m


def encode_message(msg: SendableMessage) -> Tuple[bytes, Optional[ContentType]]:
    """Encodes a message.

    Args:
        msg: The message to be encoded.

    Returns:
        A tuple containing the encoded message as bytes and the content type of the message.
    !!! note

        The above docstring is autogenerated by docstring-gen library (https://docstring-gen.airt.ai)
    """
    if msg is None:
        return b"", None

    if isinstance(msg, bytes):
        return msg, None

    if isinstance(msg, str):
        return msg.encode(), ContentTypes.text.value

    return (
        dump_json(msg).encode(),
        ContentTypes.json.value,
    )


@overload
def resolve_custom_func(
    custom_func: Optional[SyncCustomDecoder[MsgType]],
    default_func: SyncDecoder[MsgType],
) -> SyncDecoder[MsgType]:
    """Resolve a custom function.

    Args:
        custom_func: A custom function of type SyncCustomDecoder[MsgType]
        default_func: A default function of type SyncDecoder[MsgType]

    Returns:
        A resolved function of type SyncDecoder[MsgType]
    !!! note

        The above docstring is autogenerated by docstring-gen library (https://docstring-gen.airt.ai)
    """
    ...


@overload
def resolve_custom_func(
    custom_func: Optional[SyncCustomParser[MsgType]],
    default_func: SyncParser[MsgType],
) -> SyncParser[MsgType]:
    """Resolve a custom function.

    Args:
        custom_func: A custom function of type SyncCustomParser[MsgType]. Optional.
        default_func: A default function of type SyncParser[MsgType].

    Returns:
        A resolved function of type SyncParser[MsgType].
    !!! note

        The above docstring is autogenerated by docstring-gen library (https://docstring-gen.airt.ai)
    """
    ...


@overload
def resolve_custom_func(
    custom_func: Optional[AsyncCustomDecoder[MsgType]],
    default_func: AsyncDecoder[MsgType],
) -> AsyncDecoder[MsgType]:
    """Resolve a custom function.

    Args:
        custom_func: Optional custom function to be resolved.
        default_func: Default function to be used if custom function is not provided.

    Returns:
        Resolved function.
    !!! note

        The above docstring is autogenerated by docstring-gen library (https://docstring-gen.airt.ai)
    """
    ...


@overload
def resolve_custom_func(
    custom_func: Optional[AsyncCustomParser[MsgType]],
    default_func: AsyncParser[MsgType],
) -> AsyncParser[MsgType]:
    """Resolve a custom function.

    Args:
        custom_func: Optional custom function to be resolved.
        default_func: Default function to be used if custom function is not provided.

    Returns:
        Resolved function.
    !!! note

        The above docstring is autogenerated by docstring-gen library (https://docstring-gen.airt.ai)
    """
    ...


@overload
def resolve_custom_func(
    custom_func: Optional[CustomDecoder[MsgType]],
    default_func: Decoder[MsgType],
) -> Decoder[MsgType]:
    """Resolve a custom function.

    Args:
        custom_func: A custom decoder function.
        default_func: A default decoder function.

    Returns:
        A decoder function.
    !!! note

        The above docstring is autogenerated by docstring-gen library (https://docstring-gen.airt.ai)
    """
    ...


@overload
def resolve_custom_func(
    custom_func: Optional[CustomParser[MsgType]],
    default_func: Parser[MsgType],
) -> Parser[MsgType]:
    """Resolve a custom function.

    Args:
        custom_func: Optional custom function to be resolved.
        default_func: Default function to be used if custom function is not provided.

    Returns:
        Resolved function.
    !!! note

        The above docstring is autogenerated by docstring-gen library (https://docstring-gen.airt.ai)
    """
    ...


def resolve_custom_func(
    custom_func: Optional[Union[CustomDecoder[MsgType], CustomParser[MsgType]]],
    default_func: Union[Decoder[MsgType], Parser[MsgType]],
) -> Union[Decoder[MsgType], Parser[MsgType]]:
    """Resolve a custom function.

    Args:
        custom_func: Optional custom function of type CustomDecoder or CustomParser.
        default_func: Default function of type Decoder or Parser.

    Returns:
        The resolved function of type Decoder or Parser.
    !!! note

        The above docstring is autogenerated by docstring-gen library (https://docstring-gen.airt.ai)
    """
    if custom_func is None:
        return default_func

    original_params = inspect.signature(custom_func).parameters
    if len(original_params) == 1:
        return cast(Union[Decoder[MsgType], Parser[MsgType]], custom_func)

    else:
        name = tuple(original_params.items())[1][0]
        return partial(custom_func, **{name: default_func})  # type: ignore
